
/* 
 Enhanced by Dmitry Pakhomenko to initiate connection with remote SPP-aware device
 04.04.2014, Magictale Electronics

 TODO: optimise variables


 Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

 This software may be distributed and modified under the terms of the GNU
 General Public License version 2 (GPL2) as published by the Free Software
 Foundation and appearing in the file GPL2.TXT included in the packaging of
 this file. Please note that GPL2 Section 2[b] requires that all works based
 on this software must also be made publicly available under the terms of
 the GPL2 ("Copyleft").

 Contact information
 -------------------

 Kristian Lauszus, TKJ Electronics
 Web      :  http://www.tkjelectronics.com
 e-mail   :  kristianl@tkjelectronics.com
 */


#include "SPPi.h"
// To enable serial debugging see "settings.h"
//#define EXTRADEBUG // Uncomment to get even more debugging data
#define PRINTREPORT // Uncomment to print the report sent to the Arduino

/*
 * CRC (reversed crc) lookup table as calculated by the table generator in ETSI TS 101 369 V6.3.0.
 */
const uint8_t rfcomm_crc_table[256] PROGMEM = {/* reversed, 8-bit, poly=0x07 */
    0x00, 0x91, 0xE3, 0x72, 0x07, 0x96, 0xE4, 0x75, 0x0E, 0x9F, 0xED, 0x7C, 0x09, 0x98, 0xEA, 0x7B,
    0x1C, 0x8D, 0xFF, 0x6E, 0x1B, 0x8A, 0xF8, 0x69, 0x12, 0x83, 0xF1, 0x60, 0x15, 0x84, 0xF6, 0x67,
    0x38, 0xA9, 0xDB, 0x4A, 0x3F, 0xAE, 0xDC, 0x4D, 0x36, 0xA7, 0xD5, 0x44, 0x31, 0xA0, 0xD2, 0x43,
    0x24, 0xB5, 0xC7, 0x56, 0x23, 0xB2, 0xC0, 0x51, 0x2A, 0xBB, 0xC9, 0x58, 0x2D, 0xBC, 0xCE, 0x5F,
    0x70, 0xE1, 0x93, 0x02, 0x77, 0xE6, 0x94, 0x05, 0x7E, 0xEF, 0x9D, 0x0C, 0x79, 0xE8, 0x9A, 0x0B,
    0x6C, 0xFD, 0x8F, 0x1E, 0x6B, 0xFA, 0x88, 0x19, 0x62, 0xF3, 0x81, 0x10, 0x65, 0xF4, 0x86, 0x17,
    0x48, 0xD9, 0xAB, 0x3A, 0x4F, 0xDE, 0xAC, 0x3D, 0x46, 0xD7, 0xA5, 0x34, 0x41, 0xD0, 0xA2, 0x33,
    0x54, 0xC5, 0xB7, 0x26, 0x53, 0xC2, 0xB0, 0x21, 0x5A, 0xCB, 0xB9, 0x28, 0x5D, 0xCC, 0xBE, 0x2F,
    0xE0, 0x71, 0x03, 0x92, 0xE7, 0x76, 0x04, 0x95, 0xEE, 0x7F, 0x0D, 0x9C, 0xE9, 0x78, 0x0A, 0x9B,
    0xFC, 0x6D, 0x1F, 0x8E, 0xFB, 0x6A, 0x18, 0x89, 0xF2, 0x63, 0x11, 0x80, 0xF5, 0x64, 0x16, 0x87,
    0xD8, 0x49, 0x3B, 0xAA, 0xDF, 0x4E, 0x3C, 0xAD, 0xD6, 0x47, 0x35, 0xA4, 0xD1, 0x40, 0x32, 0xA3,
    0xC4, 0x55, 0x27, 0xB6, 0xC3, 0x52, 0x20, 0xB1, 0xCA, 0x5B, 0x29, 0xB8, 0xCD, 0x5C, 0x2E, 0xBF,
    0x90, 0x01, 0x73, 0xE2, 0x97, 0x06, 0x74, 0xE5, 0x9E, 0x0F, 0x7D, 0xEC, 0x99, 0x08, 0x7A, 0xEB,
    0x8C, 0x1D, 0x6F, 0xFE, 0x8B, 0x1A, 0x68, 0xF9, 0x82, 0x13, 0x61, 0xF0, 0x85, 0x14, 0x66, 0xF7,
    0xA8, 0x39, 0x4B, 0xDA, 0xAF, 0x3E, 0x4C, 0xDD, 0xA6, 0x37, 0x45, 0xD4, 0xA1, 0x30, 0x42, 0xD3,
    0xB4, 0x25, 0x57, 0xC6, 0xB3, 0x22, 0x50, 0xC1, 0xBA, 0x2B, 0x59, 0xC8, 0xBD, 0x2C, 0x5E, 0xCF
};

/*
 * "RFCOMM UUID" signature, a channel number comes immediately after it
 */
const uint8_t rfcomm_uuid_sign[6] PROGMEM = {0x35, 0x05, 0x19, 0x00, 0x03, 0x08};

SPPi::SPPi(BTD *p, const char* name, const char* pin, bool pair, uint8_t *addr) :
pBtd(p) // Pointer to BTD class instance - mandatory
{
    if (pBtd)
        pBtd->registerServiceClass(this); // Register it as a Bluetooth service

    pBtd->btdName = name;
    pBtd->btdPin = pin;
    pBtd->pairWithOtherDevice = pair;
        
    for (uint8_t i = 0; i < 6; i++)
        pBtd->remote_bdaddr[i] = addr[i];

    /* Set device cid for the SDP and RFCOMM channels */
    sdp_scid[0] = 0x50; // 0x0050
    sdp_scid[1] = 0x00;

    rfcomm_scid[0] = 0x51; // 0x0051
    rfcomm_scid[1] = 0x00;

    Reset();
}

void SPPi::Reset() {
    connected = false;
    RFCOMMConnected = false;
    SDPConnected = false;
    l2cap_sdp_state = L2CAP_SDP_WAIT;
    l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;
    sppIndex = 0;

    rfcomm_uuid_sign_idx = 0;
    rfcomm_found = false;
}

void SPPi::disconnect() {
    connected = false;
    // First the two L2CAP channels has to be disconencted and then the HCI connection
    if (RFCOMMConnected)
      pBtd->l2cap_disconnection_request(hci_handle, 0x0A, rfcomm_scid, rfcomm_dcid);
    if (RFCOMMConnected && SDPConnected)
      delay(1); // Add delay between commands
    if (SDPConnected)
      pBtd->l2cap_disconnection_request(hci_handle, 0x0B, sdp_scid, sdp_dcid);
    l2cap_sdp_state = L2CAP_DISCONNECT_RESPONSE;
}

void SPPi::ACLData(uint8_t* l2capinbuf) {

#ifdef EXTRADEBUG
    Notify(PSTR("\r\nIncoming Packet: "), 0x80);

    for (uint8_t i = 0; i < (l2capinbuf[2] + 4); i++) {
        D_PrintHex<uint8_t > (l2capinbuf[i], 0x80);
        Notify(PSTR(" "), 0x80);
    }
    Notify(PSTR("\r\n"), 0x80);
#endif

    if (((l2capinbuf[0] | (l2capinbuf[1] << 8)) == (hci_handle | 0x2000))) { // acl_handle_ok
        if ((l2capinbuf[6] | (l2capinbuf[7] << 8)) == 0x0001) { //l2cap_control - Channel ID for ACL-U
            if (l2capinbuf[8] == L2CAP_CMD_COMMAND_REJECT) {
#ifdef DEBUG_USB_HOST
                Notify(PSTR("\r\nL2CAP Command Rejected - Reason: "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[13], 0x80);
                Notify(PSTR(" "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[12], 0x80);
                Notify(PSTR(" Data: "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[17], 0x80);
                Notify(PSTR(" "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[16], 0x80);
                Notify(PSTR(" "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[15], 0x80);
                Notify(PSTR(" "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[14], 0x80);
#endif
            }else if (l2capinbuf[8] == L2CAP_CMD_CONNECTION_RESPONSE) {
                if (((l2capinbuf[16] | (l2capinbuf[17] << 8)) == 0x0000) && ((l2capinbuf[18] | (l2capinbuf[19] << 8)) == SUCCESSFUL)) { // Success
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nL2CAP Connection Response"), 0x80);
#endif
                    if (l2capinbuf[14] == sdp_scid[0] && l2capinbuf[15] == sdp_scid[1]) {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSDP Connection Response"), 0x80);
#endif
                        sdp_dcid[0] = l2capinbuf[12];
                        sdp_dcid[1] = l2capinbuf[13];

                        identifier++;
                        l2cap_sdp_state = L2CAP_SDP_CONN_RESPONSE;
                    }else if (l2capinbuf[14] == rfcomm_scid[0] && l2capinbuf[15] == rfcomm_scid[1]) {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nRFComm Connection Response"), 0x80);
#endif
                        rfcomm_dcid[0] = l2capinbuf[12];
                        rfcomm_dcid[1] = l2capinbuf[13];

                        identifier++;
                        l2cap_rfcomm_state = L2CAP_RFCOMM_CONN_RESPONSE;
                    }
                } 
            } else if(l2capinbuf[8] == L2CAP_CMD_CONFIG_RESPONSE) {
                if((l2capinbuf[16] | (l2capinbuf[17] << 8)) == 0x0000) { // Success
                    if(l2capinbuf[12] == sdp_scid[0] && l2capinbuf[13] == sdp_scid[1]) {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSDP Configuration Response Received"), 0x80);
                        Notify(PSTR("\r\nSDP Successfully Configured"), 0x80);
#endif
                        l2cap_sdp_state = L2CAP_SDP_SERVICE_SEARCH_ATTR1;
                        identifier++;
                        SDP_Service_Search_Attr(0, identifier);
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSDP Service Search Attribute Request 1 Sent"), 0x80);
#endif
                    } else if(l2capinbuf[12] == rfcomm_scid[0] && l2capinbuf[13] == rfcomm_scid[1]) {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nRFComm Configuration Response Received"), 0x80);
                        Notify(PSTR("\r\nRFComm Successfully Configured"), 0x80);
#endif
                        l2cap_sdp_state = L2CAP_RFCOMM_DONE;
                        identifier++;
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nRFComm SABM Sent"), 0x80);
#endif
                        rfcommAvailable = 0; // Reset number of bytes available
                        bytesRead = 0; // Reset number of bytes received
                        RFCOMMConnected = true;

                        //   channel direction, CR,channelType, pfBit,    data,      length
                        sendRfcomm(0, 0, (1 << 1), RFCOMM_SABM, (1 << 4), rfcommbuf, 0);
                    }
                }
            } else if (l2capinbuf[8] == L2CAP_CMD_CONFIG_REQUEST) {
                if (l2capinbuf[12] == sdp_scid[0] && l2capinbuf[13] == sdp_scid[1]) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nSDP Configuration Request Received"), 0x80);
#endif
                    identifier = l2capinbuf[9];
                    pBtd->l2cap_config_response(hci_handle, identifier, sdp_dcid);
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nSDP Configuration Response Sent"), 0x80);
#endif
                    l2cap_sdp_state = L2CAP_SDP_CONFIG_REQUEST;
                    identifier++;
                    delay(1);
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nSDP Configuration Request Sent"), 0x80);
#endif
                    pBtd->l2cap_config_request(hci_handle, identifier, sdp_dcid);

                }else if (l2capinbuf[12] == rfcomm_scid[0] && l2capinbuf[13] == rfcomm_scid[1]) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nRFComm Configuration Request Received"), 0x80);
#endif
                    identifier = l2capinbuf[9];

                    pBtd->l2cap_config_response(hci_handle, identifier, rfcomm_dcid);
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nRFComm Configuration Response Sent"), 0x80);
#endif
                    l2cap_rfcomm_state = L2CAP_RFCOMM_CONFIG_REQUEST;
                    identifier++;
                    delay(1);
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nRFComm Configuration Request Sent"), 0x80);
#endif
                    pBtd->l2cap_config_request(hci_handle, identifier, rfcomm_dcid);
                }

            } else if (l2capinbuf[8] == L2CAP_CMD_DISCONNECT_REQUEST) {
                if (l2capinbuf[12] == sdp_scid[0] && l2capinbuf[13] == sdp_scid[1]) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nDisconnect Request: SDP Channel"), 0x80);
#endif
                    identifier = l2capinbuf[9];

                    SDPConnected = false;
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nDisconnected SDP Channel"), 0x80);
#endif
                    pBtd->l2cap_disconnection_response(hci_handle, identifier, sdp_scid, sdp_dcid);
                    l2cap_sdp_state = L2CAP_SDP_WAIT;

                }else if (l2capinbuf[12] == rfcomm_scid[0] && l2capinbuf[13] == rfcomm_scid[1]) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nDisconnect Request: RFComm Channel"), 0x80);
#endif
                    identifier = l2capinbuf[9];

                    RFCOMMConnected = false;
                    connected = false;
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nDisconnected RFComm Channel"), 0x80);
#endif
                    pBtd->l2cap_disconnection_response(hci_handle, identifier, rfcomm_dcid, rfcomm_scid);
                    l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;

                }
            } else if (l2capinbuf[8] == L2CAP_CMD_DISCONNECT_RESPONSE) {
                if (l2capinbuf[12] == sdp_dcid[0] && l2capinbuf[13] == sdp_dcid[1]) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nDisconnect Response: SDP Channel"), 0x80);
#endif
                    SDPConnected = false;
                    l2cap_sdp_state = L2CAP_SDP_WAIT;

                }else if (l2capinbuf[12] == rfcomm_dcid[0] && l2capinbuf[13] == rfcomm_dcid[1]) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nDisconnect Response: RFComm Channel"), 0x80);
                    Notify(PSTR("\r\nDisconnected L2CAP Connection"), 0x80);
#endif
                    RFCOMMConnected = false;

                    pBtd->hci_disconnect(hci_handle);
                    hci_handle = -1; // Reset handle

                    l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;
                }

            } else if (l2capinbuf[8] == L2CAP_CMD_INFORMATION_REQUEST) {
#ifdef DEBUG_USB_HOST
                Notify(PSTR("\r\nInformation request"), 0x80);
#endif
                identifier = l2capinbuf[9];
                pBtd->l2cap_information_response(hci_handle, identifier, l2capinbuf[12], l2capinbuf[13]);
            }
            else {
#ifdef DEBUG_USB_HOST
                Notify(PSTR("\r\nL2CAP Unknown Signaling Command: "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[8], 0x80);
#endif
            }
        } else if (l2capinbuf[6] == sdp_scid[0] && l2capinbuf[7] == sdp_scid[1]) { // SDP
#ifdef EXTRADEBUG
            Notify(PSTR("\r\nSDP data:"), 0x80);
#endif
            if (l2capinbuf[8] == SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU) {

                if (l2cap_sdp_state == L2CAP_SDP_SERVICE_SEARCH_ATTR1){ 
#ifdef EXTRADEBUG
                    Notify(PSTR(" - SDP Service Search Attribute Response 1: "), 0x80);
#endif
                    remainingBytes = l2capinbuf[l2capinbuf[2] + 3];

                    parseAttrReply(l2capinbuf);

                    l2cap_sdp_state = L2CAP_SDP_SERVICE_SEARCH_ATTR2;
                    identifier++;
                    if (remainingBytes)
                    {
                        SDP_Service_Search_Attr(0, identifier, remainingBytes);
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSDP Service Search Attribute Request 2 Sent"), 0x80);
#endif
                    }else 
                    {
                        l2cap_sdp_state = L2CAP_SDP_DONE;

                        if (rfcomm_found)
                        {
                            pBtd->l2cap_connection_request(hci_handle, identifier, rfcomm_scid, RFCOMM_PSM);
#ifdef DEBUG_USB_HOST
                            Notify(PSTR("\r\nRFComm Connection Request Sent"), 0x80);
#endif
                            l2cap_rfcomm_state = L2CAP_RFCOMM_REQUEST;
                        }else{
#ifdef DEBUG_USB_HOST
                            Notify(PSTR("\r\nRFComm Channel Number not found"), 0x80);
#endif
                        }
                    }

                }else if (l2cap_sdp_state == L2CAP_SDP_SERVICE_SEARCH_ATTR2){
#ifdef EXTRADEBUG
                    Notify(PSTR(" - SDP Service Search Attribute Response 2: "), 0x80);
#endif

                    parseAttrReply(l2capinbuf);

                    l2cap_sdp_state = L2CAP_SDP_DONE;
                    identifier++;

                    if (rfcomm_found)
                    {
                        pBtd->l2cap_connection_request(hci_handle, identifier, rfcomm_scid, RFCOMM_PSM);
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nRFComm Connection Request Sent"), 0x80);
#endif
                        l2cap_rfcomm_state = L2CAP_RFCOMM_REQUEST;
                    }else{
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nRFComm Channel Number not found"), 0x80);
#endif
                    }
                }
            }
            else {
#ifdef EXTRADEBUG
                Notify(PSTR("\r\nUnknown PDU: "), 0x80);
                D_PrintHex<uint8_t > (l2capinbuf[8], 0x80);
#endif
            }
    
        } else if (l2capinbuf[6] == rfcomm_scid[0] && l2capinbuf[7] == rfcomm_scid[1]) { // RFCOMM

            rfcommChannel = l2capinbuf[8] & 0xF8;
            rfcommDirection = l2capinbuf[8] & 0x04;
            rfcommCommandResponse = l2capinbuf[8] & 0x02;
            rfcommChannelType = l2capinbuf[9] & 0xEF;
            rfcommPfBit = l2capinbuf[9] & 0x10;

#ifdef EXTRADEBUG
            Notify(PSTR("\r\nRFComm Channel: "), 0x80);
            D_PrintHex<uint8_t > (rfcommChannel >> 3, 0x80);
            Notify(PSTR(" Direction: "), 0x80);
            D_PrintHex<uint8_t > (rfcommDirection >> 2, 0x80);
            Notify(PSTR(" CommandResponse: "), 0x80);
            D_PrintHex<uint8_t > (rfcommCommandResponse >> 1, 0x80);
            Notify(PSTR(" ChannelType: "), 0x80);
            D_PrintHex<uint8_t > (rfcommChannelType, 0x80);
            Notify(PSTR(" PF_BIT: "), 0x80);
            D_PrintHex<uint8_t > (rfcommPfBit >> 4, 0x80);
#endif
            if (rfcommChannelType == RFCOMM_DISC) {
#ifdef DEBUG_USB_HOST
                Notify(PSTR("\r\nReceived Disconnect RFComm Command on channel: "), 0x80);
                D_PrintHex<uint8_t > (rfcommChannel, 0x80);
#endif
                connected = false;
//              sendRfcomm(rfcommChannel, rfcommDirection, rfcommCommandResponse, RFCOMM_UA, rfcommPfBit, rfcommbuf, 0x00); // UA Command
            }

            if (connected) 
            {
                /* Read the incoming message */
                if (rfcommChannelType == RFCOMM_UIH && rfcommChannel == (rfcommChannelConnection << 3)) {
                    uint8_t length = l2capinbuf[10] >> 1; // Get length
                    uint8_t offset = l2capinbuf[4] - length - 4; // Check if there is credit
                    if (checkFcs(&l2capinbuf[8], l2capinbuf[11 + length + offset])) {
                        uint8_t i = 0;
                        for (; i < length; i++) {
                            if (rfcommAvailable + i >= sizeof (rfcommDataBuffer)) {
#ifdef DEBUG_USB_HOST
                                Notify(PSTR("\r\nWarning: Buffer is full!"), 0x80);
#endif
                                break;
                            }
                            rfcommDataBuffer[rfcommAvailable + i] = l2capinbuf[11 + i + offset];
                        }
                        rfcommAvailable += i;
#ifdef EXTRADEBUG
                        Notify(PSTR("\r\nRFCOMM Data Available: "), 0x80);
                        Notify(rfcommAvailable, 0x80);
                        if (offset) {
                            Notify(PSTR(" - Credit: 0x"), 0x80);
                            D_PrintHex<uint8_t > (l2capinbuf[11], 0x80);
                        }
#endif
                    }
#ifdef DEBUG_USB_HOST
                    else
                        Notify(PSTR("\r\nError in FCS checksum!"), 0x80);
#endif
#ifdef PRINTREPORT // Uncomment "#define PRINTREPORT" to print the report send to the Arduino via Bluetooth
                    for (uint8_t i = 0; i < length; i++)
                        Notifyc(l2capinbuf[i + 11 + offset], 0x80);
#endif
                } else if (rfcommChannelType == RFCOMM_UIH && l2capinbuf[11] == BT_RFCOMM_RPN_CMD) { // UIH Remote Port Negotiation Command
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nReceived UIH Remote Port Negotiation Command"), 0x80);
#endif
/*                    rfcommbuf[0] = BT_RFCOMM_RPN_RSP; // Command
                    rfcommbuf[1] = l2capinbuf[12]; // Length and shiftet like so: length << 1 | 1
                    rfcommbuf[2] = l2capinbuf[13]; // Channel: channel << 1 | 1
                    rfcommbuf[3] = l2capinbuf[14]; // Pre difined for Bluetooth, see 5.5.3 of TS 07.10 Adaption for RFCOMM
                    rfcommbuf[4] = l2capinbuf[15]; // Priority
                    rfcommbuf[5] = l2capinbuf[16]; // Timer
                    rfcommbuf[6] = l2capinbuf[17]; // Max Fram Size LSB
                    rfcommbuf[7] = l2capinbuf[18]; // Max Fram Size MSB
                    rfcommbuf[8] = l2capinbuf[19]; // MaxRatransm.
                    rfcommbuf[9] = l2capinbuf[20]; // Number of Frames
                    sendRfcomm(rfcommChannel, rfcommDirection, 0, RFCOMM_UIH, rfcommPfBit, rfcommbuf, 0x0A); // UIH Remote Port Negotiation Response*/
                } else if (rfcommChannelType == RFCOMM_UIH && l2capinbuf[11] == BT_RFCOMM_MSC_CMD) { // UIH Modem Status Command
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nSend UIH Modem Status Response"), 0x80);
#endif
                    /*rfcommbuf[0] = BT_RFCOMM_MSC_RSP; // UIH Modem Status Response
                    rfcommbuf[1] = 2 << 1 | 1; // Length and shiftet like so: length << 1 | 1
                    rfcommbuf[2] = l2capinbuf[13]; // Channel: (1 << 0) | (1 << 1) | (0 << 2) | (channel << 3)
                    rfcommbuf[3] = l2capinbuf[14];
                    sendRfcomm(rfcommChannel, rfcommDirection, 0, RFCOMM_UIH, rfcommPfBit, rfcommbuf, 0x04);*/
                }
            }else{

                if (rfcommChannelType == RFCOMM_SABM) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nReceived SAMB RFComm Packet on channel: "), 0x80);
                    D_PrintHex<uint8_t > (rfcommChannel >> 3, 0x80);
#endif
                } else if (rfcommChannelType == RFCOMM_UA) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nReceived UA RFComm Packet on channel: "), 0x80);
                    D_PrintHex<uint8_t > (rfcommChannel >> 3, 0x80);
#endif
                    if ((rfcommChannel >> 3) == 0)
                    { 
                        //Reply on 1-st SAMB command on channel 0
                        rfcommbuf[0] = BT_RFCOMM_PN_CMD; // UIH Parameter Negotiation Request
                        rfcommbuf[1] = ((8 << 1) | 1); // Length and shiftet like so: length << 1 | 1
                        rfcommbuf[2] = ((0x10<< 1)); // Channel: channel << 1 | 1 TODO: replace channel with variable
                        rfcommbuf[3] = 0xE0; // Pre defined for Bluetooth, see 5.5.3 of TS 07.10 Adaption for RFCOMM
                        rfcommbuf[4] = 0x00; // Priority
                        rfcommbuf[5] = 0x00; // Timer
                        rfcommbuf[6] = BULK_MAXPKTSIZE - 14; // Max Fram Size LSB - set to the size of received data (50)
                        rfcommbuf[7] = 0x00; // Max Fram Size MSB
                        rfcommbuf[8] = 0x00; // MaxRetransm.
                        rfcommbuf[9] = 0x00; // Number of Frames
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSent UAH RFComm Cmd BT_RFCOMM_PN_CMD (Parameter Negotiation Request) on channel: "), 0x80);
                        D_PrintHex<uint8_t > (rfcommChannel >> 3, 0x80);
#endif
                        //   channel direction, CR,channelType,pfBit, data, length
                        sendRfcomm(0, 0, (1 << 1), RFCOMM_UIH, 0, rfcommbuf, 0x0A);
                   }else{
                        //Reply on 2-nd SAMB command on channel 0
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSend UIH RFComm Cmd BT_RFCOMM_MSC_CMD (Modem Status Command)"), 0x80);
#endif
                        rfcommbuf[0] = BT_RFCOMM_MSC_CMD; // UIH Modem Status Command
                        rfcommbuf[1] = 2 << 1 | 1; // Length and shiftet like so: length << 1 | 1
                        rfcommbuf[2] = (1 << 0) | (1 << 1) | (0 << 2) | (rfcommChannelConnection << 3); //0x83 // Channel: (1 << 0) | (1 << 1) | (0 << 2) | (channel << 3)
                        rfcommbuf[3] = 0x8D; // Can receive frames (YES), Ready to Communicate (YES), Ready to Receive (YES), Incomig Call (NO), Data is Value (YES)

                        //   channel direction, CR,channelType,pfBit, data, length
                        sendRfcomm(0, 0, (1 << 1), RFCOMM_UIH, 0, rfcommbuf, 0x04);
                   }
                } else if (rfcommChannelType == RFCOMM_UIH) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nReceived UIH RFComm Packet on channel: "), 0x80);
                    D_PrintHex<uint8_t > (rfcommChannel >> 3, 0x80);
#endif
                    if (l2capinbuf[11] == BT_RFCOMM_PN_RSP)
                    {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR(" - BT_RFCOMM_PN_RSP (Parameter Negotiation Response)"), 0x80);
                        Notify(PSTR("\r\nRFComm 2-nd SABM Sent"), 0x80);
#endif
                        //   channel direction, CR,channelType, pfBit,    data,      length
                        sendRfcomm((rfcommChannelConnection << 3), 0, (1 << 1), RFCOMM_SABM, (1 << 4), rfcommbuf, 0);
                    }else if (l2capinbuf[11] == BT_RFCOMM_MSC_CMD)
                    {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR(" - BT_RFCOMM_MSC_CMD (Modem Status Cmd)"), 0x80);
#endif
                        rfcommbuf[0] = BT_RFCOMM_MSC_RSP; // UIH Modem Status Command
                        rfcommbuf[1] = 2 << 1 | 1; // Length and shiftet like so: length << 1 | 1
                        rfcommbuf[2] = (1 << 0) | (1 << 1) | (0 << 2) | (rfcommChannelConnection << 3); //0x83 // Channel: (1 << 0) | (1 << 1) | (0 << 2) | (channel << 3)
                        rfcommbuf[3] = 0x8D; // Can receive frames (YES), Ready to Communicate (YES), Ready to Receive (YES), Incomig Call (NO), Data is Value (YES)
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nSend UIH RFComm Cmd BT_RFCOMM_MSC_RSP (Modem Status Response)"), 0x80);
#endif
                        //   channel direction, CR,channelType,pfBit, data, length
                        sendRfcomm(0, 0, (1 << 1), RFCOMM_UIH, 0, rfcommbuf, 0x04);
                    }else if (l2capinbuf[11] == BT_RFCOMM_MSC_RSP)
                    {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR(" - BT_RFCOMM_MSC_RSP (Modem Status Response)"), 0x80);
                        Notify(PSTR("\r\nRFComm Cmd with Credit Sent"), 0x80);
#endif
                        sendRfcommCredit((0x10 << 3), 0, (1 << 1), RFCOMM_UIH, 0x10, 
                            sizeof (rfcommDataBuffer)); // Send credit

                        connected = true; // The RFCOMM channel is now established
                        sppIndex = 0;
                    }else if (l2capinbuf[11] == BT_RFCOMM_RPN_CMD)
                    {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR(" - BT_RFCOMM_RPN_CMD (Remote Port Negotiation Cmd)"), 0x80);
#endif
                        //Just a stub. This command is optional according to the spec
                    }else if (l2capinbuf[11] == BT_RFCOMM_RPN_RSP)
                    {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR(" - BT_RFCOMM_RPN_RSP (Remote Port Negotiation Response)"), 0x80);
#endif
                        //Just a stub. This command is optional according to the spec
                    }else
                    {
#ifdef DEBUG_USB_HOST
                        Notify(PSTR(" - Unknown Response: "), 0x80);
                        D_PrintHex<uint8_t > (l2capinbuf[11], 0x80);
#endif
                    }
                } else if (rfcommChannelType == RFCOMM_DM) {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nReceived DM RFComm Command on channel: "), 0x80);
                    D_PrintHex<uint8_t > (rfcommChannel >> 3, 0x80);
#endif
                } else {
#ifdef DEBUG_USB_HOST
                    Notify(PSTR("\r\nUnknown Response, rfCommChannelType: "), 0x80);
                    D_PrintHex<uint8_t > (rfcommChannelType, 0x80);
#endif
                }
            }
        } else {
#ifdef EXTRADEBUG
            Notify(PSTR("\r\nUnsupported L2CAP Data - Channel ID: "), 0x80);
            D_PrintHex<uint8_t > (l2capinbuf[7], 0x80);
            Notify(PSTR(" "), 0x80);
            D_PrintHex<uint8_t > (l2capinbuf[6], 0x80);
            Notify(PSTR("\r\n"), 0x80);

            for (uint8_t i = 0; i < BULK_MAXPKTSIZE; i++) {
                D_PrintHex<uint8_t > (l2capinbuf[i], 0x80);
                Notify(PSTR(" "), 0x80);
            }
#endif
        }
    }else{
#ifdef EXTRADEBUG
        Notify(PSTR("\r\nBad ACL handle: "), 0x80);
        D_PrintHex<uint8_t > (l2capinbuf[0], 0x80);
        Notify(PSTR(" "), 0x80);
        D_PrintHex<uint8_t > (l2capinbuf[1], 0x80);
        Notify(PSTR(" "), 0x80);
#endif
    }
}

void SPPi::Run() {
    if (pBtd->pairWithOtherDevice){
        if (l2cap_sdp_state == L2CAP_SDP_WAIT){
            if (pBtd->connectToOtherDevice && !pBtd->l2capConnectionClaimed && !connected) {
                pBtd->l2capConnectionClaimed = true;
#ifdef DEBUG_USB_HOST
                Notify(PSTR("\r\nSDP Connection Request Sent"), 0x80);
#endif
                hci_handle = pBtd->hci_handle; // Store the HCI Handle for the connection
                identifier = 0;

                pBtd->l2cap_connection_request(hci_handle, identifier, sdp_scid, SDP_PSM);
                l2cap_sdp_state = L2CAP_SDP_REQUEST;
            }
        }
    }
}

/************************************************************/
/*                    SDP Commands                          */

/************************************************************/
void SPPi::SDP_Command(uint8_t* data, uint8_t nbytes) { // See page 223 in the Bluetooth specs
    pBtd->L2CAP_Command(hci_handle, data, nbytes, sdp_dcid[0], sdp_dcid[1]);
}

void SPPi::SDP_Service_Search_Attr(uint8_t transactionIDHigh, uint8_t transactionIDLow,
        uint8_t remainingLen) {

    l2capoutbuf[0] = SDP_SERVICE_SEARCH_ATTRIBUTE_REQUEST_PDU;
    l2capoutbuf[1] = transactionIDHigh;
    l2capoutbuf[2] = transactionIDLow;

    l2capoutbuf[3] = 0x00; // Parameter Length
    if (remainingLen == 0)
        l2capoutbuf[4] = 0x0F; // Parameter Length
    else
        l2capoutbuf[4] = 0x11; // Parameter Length

    l2capoutbuf[5] = 0x35; // Data Element Sequence, data size in next 8 bits
    l2capoutbuf[6] = 0x03; // Data size

    l2capoutbuf[7] = 0x19; // 00011 001 UUID, 2 bytes
    l2capoutbuf[8] = 0x01; 
    l2capoutbuf[9] = 0x00; // 0x0100 - L2CAP_UUID

    l2capoutbuf[10] = 0x00;
    l2capoutbuf[11] = 0x26;// 0x0026 Maximum attribute byte count

    l2capoutbuf[12] = 0x35;// Data Element Sequence, data size in next 8 bits
    l2capoutbuf[13] = 0x05;// Data size

    l2capoutbuf[14] = 0x0A;// 00001 010 unsigned int, 4 bytes  - Attribute ID range

    l2capoutbuf[15] = 0x00;
    l2capoutbuf[16] = 0x00;// range from 0x0000...

    l2capoutbuf[17] = 0xFF;
    l2capoutbuf[18] = 0xFF;// ... to 0xFFFF

    if (remainingLen == 0)
    {       
        l2capoutbuf[19] = 0x00;// No more data
        SDP_Command(l2capoutbuf, 20);          
    }else{
        l2capoutbuf[19] = 0x02;
        l2capoutbuf[20] = 0x0; //will be 0 anyway
        l2capoutbuf[21] = remainingLen;
        SDP_Command(l2capoutbuf, 22);
    }
}

/************************************************************/
/*                    RFCOMM Commands                       */

/************************************************************/
void SPPi::RFCOMM_Command(uint8_t* data, uint8_t nbytes) {
    pBtd->L2CAP_Command(hci_handle, data, nbytes, rfcomm_dcid[0], rfcomm_dcid[1]);
}

void SPPi::sendRfcomm(uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t* data, uint8_t length) {
    l2capoutbuf[0] = channel | direction | CR | extendAddress; // RFCOMM Address
    l2capoutbuf[1] = channelType | pfBit; // RFCOMM Control
    l2capoutbuf[2] = length << 1 | 0x01; // Length and format (always 0x01 bytes format)
    uint8_t i = 0;
    for (; i < length; i++)
        l2capoutbuf[i + 3] = data[i];
    l2capoutbuf[i + 3] = calcFcs(l2capoutbuf);
#ifdef EXTRADEBUG
    Notify(PSTR(" - RFComm Data: "), 0x80);
    for (i = 0; i < length + 4; i++) {
        D_PrintHex<uint8_t > (l2capoutbuf[i], 0x80);
        Notify(PSTR(" "), 0x80);
    }
#endif
    RFCOMM_Command(l2capoutbuf, length + 4);
}

void SPPi::sendRfcommCredit(uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t credit) {
    l2capoutbuf[0] = channel | direction | CR | extendAddress; // RFCOMM Address
    l2capoutbuf[1] = channelType | pfBit; // RFCOMM Control
    l2capoutbuf[2] = 0x01; // Length = 0
    l2capoutbuf[3] = credit; // Credit
    l2capoutbuf[4] = calcFcs(l2capoutbuf);
#ifdef EXTRADEBUG
    Notify(PSTR(" - RFCOMM Credit Data: "), 0x80);
    for (uint8_t i = 0; i < 5; i++) {
        D_PrintHex<uint8_t > (l2capoutbuf[i], 0x80);
        Notify(PSTR(" "), 0x80);
    }
#endif
    RFCOMM_Command(l2capoutbuf, 5);
}

/* CRC on 2 bytes */
uint8_t SPPi::crc(uint8_t *data) {
    return (pgm_read_byte(&rfcomm_crc_table[pgm_read_byte(&rfcomm_crc_table[0xFF ^ data[0]]) ^ data[1]]));
}

/* Calculate FCS */
uint8_t SPPi::calcFcs(uint8_t *data) {
    if ((data[1] & 0xEF) == RFCOMM_UIH)
        return (0xFF - crc(data)); // FCS on 2 bytes
    else
        return (0xFF - pgm_read_byte(&rfcomm_crc_table[crc(data) ^ data[2]])); // FCS on 3 bytes
}

/* Check FCS */
bool SPPi::checkFcs(uint8_t *data, uint8_t fcs) {
    uint8_t temp = crc(data);
    if ((data[1] & 0xEF) != RFCOMM_UIH)
        temp = pgm_read_byte(&rfcomm_crc_table[temp ^ data[2]]); // FCS on 3 bytes
    return (pgm_read_byte(&rfcomm_crc_table[temp ^ fcs]) == 0xCF);
}

/* Serial commands */
size_t SPPi::write(uint8_t data) {
    return write(&data, 1);
}

size_t SPPi::write(const uint8_t *data, size_t size) {
    for(uint8_t i = 0; i < size; i++) {
        if(sppIndex >= sizeof(sppOutputBuffer)/sizeof(sppOutputBuffer[0]))
            send(); // Send the current data in the buffer
        sppOutputBuffer[sppIndex++] = data[i]; // All the bytes are put into a buffer and then send using the send() function
    }
    return size;
}

void SPPi::send(void)
{
    if (!connected || !sppIndex) return;

    uint8_t length; // This is the length of the string we are sending
    uint8_t offset = 0; // This is used to keep track of where we are in the string

    l2capoutbuf[0] = (rfcommChannelConnection << 3)| 0 | 0 | extendAddress; // RFCOMM Address
    l2capoutbuf[1] = RFCOMM_UIH; // RFCOMM Control

    while (sppIndex) { // We will run this while loop until this variable is 0
        if (sppIndex > (sizeof (l2capoutbuf) - 4)) // Check if the string is larger than the outgoing buffer
            length = sizeof (l2capoutbuf) - 4;
        else
            length = sppIndex;

        l2capoutbuf[2] = length << 1 | 1; // Length
        uint8_t i = 0;
        for (; i < length; i++)
            l2capoutbuf[i + 3] = sppOutputBuffer[i + offset];
        l2capoutbuf[i + 3] = calcFcs(l2capoutbuf); // Calculate checksum

        RFCOMM_Command(l2capoutbuf, length + 4);

        sppIndex -= length;
        offset += length; // Increment the offset
    }
}

int SPPi::available(void)
{
    return rfcommAvailable;
}

int SPPi::peek(void)
{
    if (rfcommAvailable == 0) // Don't read if there is nothing in the buffer
        return -1;
    return rfcommDataBuffer[0];
}

int SPPi::read(void)
{
    if (rfcommAvailable == 0) // Don't read if there is nothing in the buffer
        return -1;
    uint8_t output = rfcommDataBuffer[0];
    for (uint8_t i = 1; i < rfcommAvailable; i++)
        rfcommDataBuffer[i - 1] = rfcommDataBuffer[i]; // Shift the buffer one left
    rfcommAvailable--;
    bytesRead++;
    if (bytesRead > (sizeof (rfcommDataBuffer) - 5)) { // We will send the command just before it runs out of credit
        bytesRead = 0;
        sendRfcommCredit((rfcommChannelConnection << 3), rfcommDirection, 0, RFCOMM_UIH, 0x10, sizeof (rfcommDataBuffer)); // Send more credit
#ifdef EXTRADEBUG
        Notify(PSTR("\r\nSent "), 0x80);
        Notify((uint8_t)sizeof (rfcommDataBuffer), 0x80);
        Notify(PSTR(" more credit"), 0x80);
#endif
    }
    return output;
}

void SPPi::parseAttrReply(uint8_t* l2capinbuf)
{
    if ((l2capinbuf[2] + 4) < 15) return; //sanity check
    
    if (rfcomm_found) 
    {
#ifdef DEBUG_USB_HOST
        Notify(PSTR("\r\nChannel is already found"), 0x80);
#endif
        return;
    }

    if (rfcomm_uuid_sign_idx == sizeof(rfcomm_uuid_sign))
    {
        //Signature has been already found but channel is in next packet
        rfcommChannelConnection = l2capinbuf[15];
        rfcomm_found = true;

#ifdef DEBUG_USB_HOST
        Notify(PSTR("\r\nChannel found in second packet: "), 0x80);
        Notify((uint8_t)rfcommChannelConnection, 0x80);
#endif
        return;
    }

    uint8_t nextSignBt;
    for (uint8_t i = 15; i < (l2capinbuf[2] + 4); i++) //searching through packet payload only
    {
        //Keep searching for signature
        if (l2capinbuf[i] == pgm_read_byte(&rfcomm_uuid_sign[rfcomm_uuid_sign_idx]))
        {
#ifdef EXTRADEBUG
            Notify(PSTR("\r\nFound match @"), 0x80);      
            Notify((uint8_t)i, 0x80);
#endif
            rfcomm_uuid_sign_idx++;
            if (rfcomm_uuid_sign_idx == sizeof(rfcomm_uuid_sign))
            {
                //signature found, trying to get channel number
                if (l2capinbuf[i + 1] == 0x2)
                {
                    //is the byte we are looking at the second last in the packet?
                    if ((l2capinbuf[2] + 1) ==  (i + 1))
                    {
                        //Oh well, channel number didn't fit in this packet, waiting for next
#ifdef DEBUG_USB_HOST
                        Notify(PSTR("\r\nChannel will be in the next packet"), 0x80);      
#endif
                        return;
                    }
                }

                rfcommChannelConnection = l2capinbuf[i + 1];
                rfcomm_found = true;
#ifdef DEBUG_USB_HOST
                Notify(PSTR("\r\nChannel found in first packet: "), 0x80);
                D_PrintHex<uint8_t > (rfcommChannelConnection, 0x80);
#endif
                return;
            }
        }else if ((l2capinbuf[i] == 0x2) && (i == l2capinbuf[2] + 1))
        {
            //This is an indication of packet end with more data to come in next packet - 
            //do not reset rfcomm_uuid_sign_idx, we will continue when next packet arrives
#ifdef DEBUG_USB_HOST
            Notify(PSTR("\r\nEnd of packet reached, no full signature yet"), 0x80);
#endif
            return; 
        }else
        {  
            //Otherwise the signature is not found - start over again
            rfcomm_uuid_sign_idx = 0;
        }
    }
    return;
}


